"""Flow for building a ubuntu classic disk image."""

import os
import re
import shutil
import logging

from subprocess import CalledProcessError
from tempfile import gettempdir
from ubuntu_image.common_builder import AbstractImageBuilderState
from ubuntu_image.helpers import (
     check_root_privilege, get_host_arch, live_build, run)


DEFAULT_FS = 'ext4'
DEFAULT_FS_LABEL = 'writable'
_logger = logging.getLogger('ubuntu-image')


class ClassicBuilder(AbstractImageBuilderState):
    def __init__(self, args):
        super().__init__(args)
        self.gadget_tree = args.gadget_tree
        # It's required to run ubuntu-image as root to build classic image.
        check_root_privilege()

    def __getstate__(self):
        state = super().__getstate__()
        state.update(
            gadget_tree=self.gadget_tree,
            )
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        self.gadget_tree = state['gadget_tree']

    def prepare_gadget_tree(self):
        gadget_dir = os.path.join(self.unpackdir, 'gadget')
        shutil.copytree(self.gadget_tree, gadget_dir)
        # We assume the gadget tree was built from a gadget source tree using
        # snapcraft prime so the gadget.yaml file is expected in the meta/
        # directory.
        self.yaml_file_path = os.path.join(
            gadget_dir, 'meta', 'gadget.yaml')
        super().prepare_gadget_tree()

    def prepare_image(self):
        if not self.args.filesystem:
            try:
                # Configure it with environment variables.
                env = {}
                if self.args.project is not None:
                    env['PROJECT'] = self.args.project
                if self.args.suite is not None:
                    env['SUITE'] = self.args.suite
                if self.args.arch is not None:
                    env['ARCH'] = self.args.arch
                if self.args.subproject is not None:
                    env['SUBPROJECT'] = self.args.subproject
                if self.args.subarch is not None:
                    env['SUBARCH'] = self.args.subarch
                if self.args.with_proposed:
                    env['PROPOSED'] = '1'
                if self.args.extra_ppas is not None:
                    env['EXTRA_PPAS'] = ' '.join(self.args.extra_ppas)
                # Only generate a single rootfs tree for classic images.
                env['IMAGEFORMAT'] = 'none'
                # ensure ARCH is set
                if self.args.arch is None:
                    env['ARCH'] = get_host_arch()
                live_build(self.unpackdir, env)
            except CalledProcessError:
                if self.args.debug:
                    _logger.exception('Full debug traceback follows')
                self.exitcode = 1
                # Stop the state machine here by not appending a next step.
                return

        super().prepare_image()

    def populate_rootfs_contents(self):
        dst = self.rootfs
        if self.args.filesystem:
            src = self.args.filesystem
            # 'cp -a' is faster than the python functions and makes sure all
            # meta information is preserved.
            run('cp -a {} {}'.format(os.path.join(src, '*'), dst), shell=True)
        else:
            src = os.path.join(self.unpackdir, 'chroot')
            for subdir in os.listdir(src):
                shutil.move(os.path.join(src, subdir),
                            os.path.join(dst, subdir))
        # Replace pre-defined LABEL in /etc/fstab with the one
        # we're using 'LABEL=writable' in grub.cfg.
        # TODO We need EFI partition in fstab too
        fstab_path = os.path.join(dst, 'etc', 'fstab')
        if os.path.exists(fstab_path):
            with open(fstab_path, 'r') as fstab:
                new_content = re.sub(r'(LABEL=)\S+',
                                     r'\1{}'.format(DEFAULT_FS_LABEL),
                                     fstab.read(), count=1)
            # Insert LABEL entry if it's not found at fstab
            fs_label = 'LABEL={}'.format(DEFAULT_FS_LABEL)
            if fs_label not in new_content:
                new_content += 'LABEL={}   /    {}   defaults    0 0'.format(
                       DEFAULT_FS_LABEL, DEFAULT_FS)
            with open(fstab_path, 'w') as fstab:
                fstab.write(new_content)
        if self.cloud_init is not None:
            # LP: #1633232 - Only write out meta-data when the --cloud-init
            # parameter is given.
            seed_dir = os.path.join(dst, 'var', 'lib', 'cloud', 'seed')
            cloud_dir = os.path.join(seed_dir, 'nocloud-net')
            os.makedirs(cloud_dir, exist_ok=True)
            metadata_file = os.path.join(cloud_dir, 'meta-data')
            with open(metadata_file, 'w', encoding='utf-8') as fp:
                print('instance-id: nocloud-static', file=fp)
            userdata_file = os.path.join(cloud_dir, 'user-data')
            shutil.copy(self.cloud_init, userdata_file)
        super().populate_rootfs_contents()

    def generate_manifests(self):
        # After the images are built, we would also like to have some image
        # manifests exported so that one can easily check what packages have
        # been installed on the rootfs. We utilize dpkg-query tool to generate
        # the manifest file for classic image. Packages like casper which is
        # only useful in a live CD/DVD are removed.
        # The deprecated words can be found below:
        # https://help.ubuntu.com/community/MakeALiveCD/DVD/BootableFlashFromHarddiskInstall
        deprecated_words = ['ubiquity', 'casper']
        manifest_path = os.path.join(self.output_dir, 'filesystem.manifest')
        tmpfile_path = os.path.join(gettempdir(), 'filesystem.manifest')
        with open(tmpfile_path, 'w+') as tmpfile:
            query_cmd = ['sudo', 'chroot', self.rootfs, 'dpkg-query', '-W',
                         '--showformat=${Package} ${Version}\n']
            run(query_cmd, stdout=tmpfile, stderr=None, env=os.environ)
            tmpfile.seek(0, 0)
            with open(manifest_path, 'w') as manifest:
                for line in tmpfile:
                    if not any(word in line for word in deprecated_words):
                        manifest.write(line)
        super().generate_manifests()



"""Test the hook mechanism."""

import os

from contextlib import ExitStack
from tempfile import TemporaryDirectory
from textwrap import dedent
from ubuntu_image.hooks import HookError, HookManager
from unittest import TestCase


class TestHooks(TestCase):
    def test_hook_compatibility(self):
        # This test should be updated whenever NEW hooks are added.  It is NOT
        # allowed to remove any hooks from this test - it's present here to
        # make sure no existing hooks have been
        pass

    def test_hook_fired(self):
        with ExitStack() as resources:
            hooksdir = resources.enter_context(TemporaryDirectory())
            hookfile = os.path.join(hooksdir, 'test-hook')
            resultfile = os.path.join(hooksdir, 'result')
            env = {'UBUNTU_IMAGE_TEST_ENV': 'true'}
            with open(hookfile, 'w') as fp:
                fp.write("""\
#!/bin/sh
echo -n "$UBUNTU_IMAGE_TEST_ENV" >>{}
""".format(resultfile))
            os.chmod(hookfile, 0o744)
            manager = HookManager([hooksdir])
            manager.fire('test-hook', env)
            # Check if the script ran once as expected.
            self.assertTrue(os.path.exists(resultfile))
            with open(resultfile, 'r') as fp:
                self.assertEqual(fp.read(), 'true')

    def test_hook_fired_multiple_scripts(self):
        with ExitStack() as resources:
            hooksdir = resources.enter_context(TemporaryDirectory())
            hookdir = os.path.join(hooksdir, 'test-hook.d')
            hookfile1 = os.path.join(hookdir, 'dir-test-01')
            hookfile2 = os.path.join(hookdir, 'dir-test-02')
            hookfile3 = os.path.join(hooksdir, 'test-hook')
            resultfile = os.path.join(hooksdir, 'result')
            os.mkdir(hookdir)

            def create_hook(path):
                with open(path, 'w') as fp:
                    fp.write(dedent("""\
                                    #!/bin/sh
                                    echo "{}" >>{}
                                    """.format(path, resultfile)))
                os.chmod(path, 0o744)
            create_hook(hookfile1)
            create_hook(hookfile2)
            create_hook(hookfile3)
            manager = HookManager([hooksdir])
            manager.fire('test-hook')
            # Check if all the scripts for the hook were run and in the right
            # order.
            self.assertTrue(os.path.exists(resultfile))
            with open(resultfile, 'r') as fp:
                lines = fp.read().splitlines()
                self.assertListEqual(
                    lines, [hookfile1, hookfile2, hookfile3])

    def test_hook_multiple_directories(self):
        with ExitStack() as resources:
            hooksdir1 = resources.enter_context(TemporaryDirectory())
            hooksdir2 = resources.enter_context(TemporaryDirectory())
            hookdir = os.path.join(hooksdir1, 'test-hook.d')
            hookfile1 = os.path.join(hookdir, 'dir-test-01')
            hookfile2 = os.path.join(hooksdir2, 'test-hook')
            # We write the results to one file to check if order is proper.
            resultfile = os.path.join(hooksdir1, 'result')
            os.mkdir(hookdir)

            def create_hook(path):
                with open(path, 'w') as fp:
                    fp.write(dedent("""\
                                    #!/bin/sh
                                    echo "{}" >>{}
                                    """.format(path, resultfile)))
                os.chmod(path, 0o744)
            create_hook(hookfile1)
            create_hook(hookfile2)
            manager = HookManager([hooksdir1, hooksdir2])
            manager.fire('test-hook')
            # Check if all the scripts for the hook were run and in the right
            # order.
            self.assertTrue(os.path.exists(resultfile))
            with open(resultfile, 'r') as fp:
                lines = fp.read().splitlines()
                self.assertListEqual(
                    lines, [hookfile1, hookfile2])

    def test_hook_error(self):
        with ExitStack() as resources:
            hooksdir = resources.enter_context(TemporaryDirectory())
            hookfile = os.path.join(hooksdir, 'test-hook')
            with open(hookfile, 'w') as fp:
                fp.write(dedent("""\
                                #!/bin/sh
                                echo -n "error" 1>&2
                                exit 1
                                """))
            os.chmod(hookfile, 0o744)
            manager = HookManager([hooksdir])
            # Check if hook script failures are properly reported
            with self.assertRaises(HookError) as cm:
                manager.fire('test-hook')
            self.assertEqual(cm.exception.hook_name, 'test-hook')
            self.assertEqual(cm.exception.hook_path, hookfile)
            self.assertEqual(cm.exception.hook_retcode, 1)
            self.assertEqual(cm.exception.hook_stderr, 'error')



"""Flow for building a ubuntu core image."""

import os
import shutil
import logging

from pathlib import Path
from subprocess import CalledProcessError
from ubuntu_image.common_builder import AbstractImageBuilderState
from ubuntu_image.helpers import snap


_logger = logging.getLogger('ubuntu-image')


class ModelAssertionBuilder(AbstractImageBuilderState):
    def __init__(self, args):
        super().__init__(args)

    def prepare_image(self):
        # Since some people might still use the deprecated extra-snaps syntax,
        # combine the two argument lists before sending it out to
        # prepare-image.
        extra_snaps = (self.args.snap or []) + (self.args.extra_snaps or [])
        try:
            snap(self.args.model_assertion, self.unpackdir,
                 self.args.channel, extra_snaps)
        except CalledProcessError:
            if self.args.debug:
                _logger.exception('Full debug traceback follows')
            self.exitcode = 1
            # Stop the state machine right here by not appending a next step.
        else:
            self.yaml_file_path = os.path.join(
                self.unpackdir, 'gadget', 'meta', 'gadget.yaml')
            super().prepare_image()

    def populate_rootfs_contents(self):
        if self.gadget.seeded:
            # For now, since we only create the system-seed partition for
            # uc20 images, we hard-code to use this path for the rootfs
            # seed population.  In the future we might want to consider
            # populating other partitions from `snap prepare-image` output
            # as well, so looking into directories like system-data/ etc.
            src = os.path.join(self.unpackdir, 'system-seed')
            dst = self.rootfs
        else:
            src = os.path.join(self.unpackdir, 'image')
            dst = os.path.join(self.rootfs, 'system-data')
            # This is just a mount point.
            os.makedirs(os.path.join(dst, 'boot'), exist_ok=True)
        for subdir in os.listdir(src):
            # LP: #1632134 - copy everything under the image directory except
            # /boot which goes to the boot partition. Unless this is a uc20
            # system which has everything under "system-seed".
            if not self.gadget.seeded and subdir == 'boot':
                continue
            shutil.move(os.path.join(src, subdir),
                        os.path.join(dst, subdir))
        etc_cloud = os.path.join(dst, 'etc', 'cloud')
        if os.path.isdir(etc_cloud) and not os.listdir(etc_cloud):
            # The snap --prepare-image command creates /etc/cloud even if
            # it's empty.  We don't want to copy it over into the final rootfs
            # in that case as it can cause issues when base snaps want to ship
            # default configuration there.
            os.rmdir(etc_cloud)
        if self.cloud_init is not None:
            # LP: #1633232 - Only write out meta-data when the --cloud-init
            # parameter is given.
            seed_dir = os.path.join(dst, 'var', 'lib', 'cloud', 'seed')
            cloud_dir = os.path.join(seed_dir, 'nocloud-net')
            os.makedirs(cloud_dir, exist_ok=True)
            metadata_file = os.path.join(cloud_dir, 'meta-data')
            with open(metadata_file, 'w', encoding='utf-8') as fp:
                print('instance-id: nocloud-static', file=fp)
            userdata_file = os.path.join(cloud_dir, 'user-data')
            shutil.copy(self.cloud_init, userdata_file)
        if self.disable_console_conf:
            # For now we just touch /var/lib/console-conf/complete to disable
            # console-conf on core images.
            cc_dir = os.path.join(dst, 'var', 'lib', 'console-conf')
            os.makedirs(cc_dir, exist_ok=True)
            Path(os.path.join(cc_dir, 'complete')).touch()
        super().populate_rootfs_contents()

    def _write_manifest(self, snaps_dir, filename):
        if os.path.isdir(snaps_dir):
            manifest_path = os.path.join(self.output_dir, filename)
            with open(manifest_path, 'w') as manifest:
                for file in os.listdir(snaps_dir):
                    if file.endswith('.snap'):
                        parts = file[:-5].rpartition('_')
                        manifest.write('{} {}\n'.format(parts[0], parts[2]))

    def generate_manifests(self):
        # After the images are built, we would also like to have some image
        # manifests exported so that one can easily check what snap packages
        # have been installed as part of the image.
        # We generate two files - one based off the snaps/ directory and other
        # basing on the contents of seed/snaps.
        # snaps.manifest
        snaps_dir = os.path.join(
            self.rootfs, 'system-data', 'var', 'lib', 'snapd', 'snaps')
        self._write_manifest(snaps_dir, 'snaps.manifest')
        # seed.manifest
        if self.gadget.seeded:
            # For uc20 builds, the snaps are located in a different directory.
            seed_dir = os.path.join(
                self.rootfs, 'snaps')
        else:
            seed_dir = os.path.join(
                self.rootfs, 'system-data', 'var', 'lib', 'snapd', 'seed',
                'snaps')
        self._write_manifest(seed_dir, 'seed.manifest')
        super().generate_manifests()



"""Testing helpers."""

import os
import shutil
import logging

from contextlib import ExitStack, contextmanager
from pkg_resources import resource_filename
from subprocess import PIPE, run as subprocess_run
from types import SimpleNamespace
from ubuntu_image.assertion_builder import ModelAssertionBuilder
from ubuntu_image.classic_builder import ClassicBuilder
from unittest.mock import patch


DIRS_UNDER_ROOTFS = ['bin', 'boot', 'dev', 'etc', 'home', 'lib',
                     'lib64', 'media', 'initrd.img', 'mnt', 'opt',
                     'proc', 'root', 'run', 'sbin', 'snap', 'srv',
                     'sys', 'tmp', 'usr', 'var', 'vmlinuz']


class XXXModelAssertionBuilder(ModelAssertionBuilder):
    exitcode = 0
    gadget_yaml = 'gadget.yaml'

    # We need this class because the current gadget snap we get from the store
    # does not contain a gadget.yaml or grub files, although it (probably)
    # will eventually.  For now, this copies sample files into the expected
    # case, and should be used in tests which require that step.
    def load_gadget_yaml(self):
        gadget_dir = os.path.join(self.unpackdir, 'gadget')
        meta_dir = os.path.join(gadget_dir, 'meta')
        self.yaml_file_path = os.path.join(meta_dir, 'gadget.yaml')
        os.makedirs(meta_dir, exist_ok=True)
        shutil.copy(
            resource_filename('ubuntu_image.tests.data', self.gadget_yaml),
            os.path.join(meta_dir, 'gadget.yaml'))
        shutil.copy(
            resource_filename('ubuntu_image.tests.data', 'grubx64.efi'),
            os.path.join(gadget_dir, 'grubx64.efi'))
        shutil.copy(
            resource_filename('ubuntu_image.tests.data', 'shim.efi.signed'),
            os.path.join(gadget_dir, 'shim.efi.signed'))
        super().load_gadget_yaml()


class XXXClassicBuilder(ClassicBuilder):
    pass


class CrashingModelAssertionBuilder(XXXModelAssertionBuilder):
    def make_temporary_directories(self):
        raise RuntimeError


class EarlyExitModelAssertionBuilder(XXXModelAssertionBuilder):
    def prepare_image(self):
        # Do nothing, but let the state machine exit.
        pass


class DoNothingBuilder(XXXModelAssertionBuilder):
    def prepare_image(self):
        self._next.append(self.load_gadget_yaml)

    def populate_rootfs_contents(self):
        self._next.append(self.populate_rootfs_contents_hooks)

    def populate_bootfs_contents(self):
        self._next.append(self.prepare_filesystems)


class EarlyExitLeaveATraceAssertionBuilder(XXXModelAssertionBuilder):
    def prepare_image(self):
        # Similar to above, but leave a trace that this method ran, so that we
        # have something to positively test.
        with open(os.path.join(self.workdir, 'success'), 'w'):
            pass


class EarlyExitLeaveATraceClassicBuilder(XXXClassicBuilder):
    def prepare_gadget_tree(self):
        # We're skipping the gadget tree build as we're leaving early and will
        # not use it for the tests.
        self._next.append(self.prepare_image)

    def prepare_image(self):
        # Similar to above, but leave a trace that this method ran, so that we
        # have something to positively test.
        with open(os.path.join(self.workdir, 'success'), 'w'):
            pass


class CallLBLeaveATraceClassicBuilder(XXXClassicBuilder):
    def prepare_gadget_tree(self):
        # We're skipping the gadget tree build as we're leaving early and will
        # not use it for the tests.
        self._next.append(self.prepare_image)

    def load_gadget_yaml(self):
        # This time we want to call prepare_image for the live-build call but
        # then finish after leaving a trace
        with open(os.path.join(self.workdir, 'success'), 'w'):
            pass


class LogCapture:
    def __init__(self):
        self.logs = []
        self._resources = ExitStack()

    def capture(self, *args, **kws):
        level, fmt, fmt_args = args
        self.logs.append((level, fmt % fmt_args))
        # Was .exception() called?
        exc_info = kws.pop('exc_info', None)
        assert len(kws) == 0, kws
        if exc_info:
            self.logs.append('IMAGINE THE TRACEBACK HERE')

    def __enter__(self):
        log = logging.getLogger('ubuntu-image')
        self._resources.enter_context(patch.object(log, '_log', self.capture))
        return self

    def __exit__(self, *exception):
        self._resources.close()
        # Don't suppress any exceptions.
        return False


class LiveBuildMocker:
    def __init__(self, root_dir):
        self.call_args_list = []
        self.root_dir = root_dir

    def run(self, command, *args, **kws):
        cmd_str = command if isinstance(command, str) else ' '.join(command)
        if 'lb config' in cmd_str:
            self.call_args_list.append(command)
            return SimpleNamespace(returncode=1)
        elif 'lb build' in cmd_str:
            self.call_args_list.append(command)
            # Create dummy top-level filesystem layout.
            chroot_dir = os.path.join(self.root_dir, 'chroot')
            for dir_name in DIRS_UNDER_ROOTFS:
                os.makedirs(os.path.join(chroot_dir, dir_name))
            return SimpleNamespace(returncode=1)
        elif cmd_str.startswith('dpkg -L'):
            self.call_args_list.append(command)
            stdout = kws.pop('stdout', PIPE)
            stderr = kws.pop('stderr', PIPE)
            return subprocess_run(
                command,
                stdout=stdout, stderr=stderr,
                universal_newlines=True,
                **kws)
        elif cmd_str.startswith('dpkg --print-architecture'):
            self.call_args_list.append(command)
            return SimpleNamespace(stdout='amd64', returncode=0)


@contextmanager
def envar(key, value):
    missing = object()
    # Temporarily set an environment variable.
    old_value = os.environ.get(key, missing)
    os.environ[key] = value
    try:
        yield
    finally:
        if old_value is missing:
            del os.environ[key]
        else:
            os.environ[key] = old_value



import os

from pkg_resources import resource_string as resource_bytes


# Try to get the version number, which will be different if we're living in a
# snap world or a deb.  Actually, I'd prefer to not even have the -NubuntuY
# version string when we're running from source, but that's trickier, so don't
# worry about it.
__version__ = os.environ.get('SNAP_VERSION')
if __version__ is None:                                      # pragma: nocover
    try:
        __version__ = resource_bytes(
            'ubuntu_image', 'version.txt').decode('utf-8')
    except FileNotFoundError:
        # Probably, setup.py hasn't been run yet to generate the version.txt.
        __version__ = 'dev'



"""Test image building."""

import os

from contextlib import suppress
from parted import IOException
from struct import unpack
from tempfile import TemporaryDirectory
from ubuntu_image.helpers import GiB, MiB
from ubuntu_image.image import Image
from ubuntu_image.parser import VolumeSchema
from unittest import TestCase


class TestImage(TestCase):
    def setUp(self):
        actual_tmpdir = TemporaryDirectory()
        self.tmpdir = actual_tmpdir.name
        self.addCleanup(actual_tmpdir.cleanup)
        self.img = os.path.join(self.tmpdir, 'img')
        assert not os.path.exists(self.img)

    def test_initialize(self):
        image = Image(self.img, GiB(1.25))
        self.assertTrue(os.path.exists(image.path))
        # GiB == 1024**3; 1.25GiB == 1342177280 bytes.
        self.assertEqual(os.stat(image.path).st_size, 1342177280)

    def test_initialize_smaller(self):
        image = Image(self.img, MiB(4.5))
        self.assertTrue(os.path.exists(image.path))
        # MiB == 1024**2; 4.5MiB == 4718592 bytes.
        self.assertEqual(os.stat(image.path).st_size, 4718592)

    def test_initialize_partition_table_gpt(self):
        image = Image(self.img, MiB(10), VolumeSchema.gpt)
        self.assertTrue(os.path.exists(image.path))
        self.assertEqual(os.stat(image.path).st_size, 10485760)
        self.assertEqual(image.disk.type, 'gpt')
        self.assertEqual(image.schema, VolumeSchema.gpt)

    def test_initialize_partition_table_mbr(self):
        image = Image(self.img, MiB(10), VolumeSchema.mbr)
        self.assertTrue(os.path.exists(image.path))
        self.assertEqual(os.stat(image.path).st_size, 10485760)
        self.assertEqual(image.disk.type, 'msdos')
        self.assertEqual(image.schema, VolumeSchema.mbr)

    def test_initialize_partition_table_mbr_too_small(self):
        # Creating a super small image should fail as there's no place for
        # a partition table.
        self.assertRaises(IOException,
                          Image, self.img, 25, VolumeSchema.gpt)
        # ...but we can create an Image object without the partition table.
        image = Image(self.img, 25, None)
        self.assertTrue(os.path.exists(image.path))
        self.assertEqual(os.stat(image.path).st_size, 25)

    def test_copy_blob_install_grub_to_mbr(self):
        # Install GRUB to MBR
        # TODO: this has to be represented in the gadget.yaml
        # NOTE: the boot.img has to be a part of the gadget snap itself
        # FIXME: embed a pointer to 2nd stage in bios-boot partition
        #
        # dd if=blobs/img.mbr of=img bs=446 count=1 conv=notrunc
        #
        # Start by creating a blob of the requested size.
        blob_file = os.path.join(self.tmpdir, 'mbr.blob')
        with open(blob_file, 'wb') as fp:
            fp.write(b'happyhappyjoyjoy' * 27)
            fp.write(b'happyhappyjoyj')
        self.assertEqual(os.stat(blob_file).st_size, 446)
        image = Image(self.img, MiB(1))
        image.copy_blob(blob_file, bs=446, count=1, conv='notrunc')
        # At the top of the image file, there should be 27 Stimpy
        # Exclamations, followed by a happyhappyjoyj.
        with open(image.path, 'rb') as fp:
            complete_stimpys = fp.read(432)
            partial_stimpys = fp.read(14)
            # Spot check.
            zeros = fp.read(108)
        self.assertEqual(complete_stimpys, b'happyhappyjoyjoy' * 27)
        self.assertEqual(partial_stimpys, b'happyhappyjoyj')
        # Stevens $4.13 - the extended file should read as zeros.
        self.assertEqual(zeros, b'\0' * 108)

    def test_copy_blob_with_seek(self):
        # dd if=blobs/img.bios-boot of=img bs=1MiB seek=4 count=1 conv=notrunc
        blob_file = os.path.join(self.tmpdir, 'img.bios-boot')
        with open(blob_file, 'wb') as fp:
            fp.write(b'x' * 100)
        image = Image(self.img, MiB(2))
        image.copy_blob(blob_file, bs=773, seek=4, count=1, conv='notrunc')
        # The seek=4 skipped 4 blocks of 773 bytes.
        with open(image.path, 'rb') as fp:
            self.assertEqual(fp.read(3092), b'\0' * 3092)
            self.assertEqual(fp.read(100), b'x' * 100)
            self.assertEqual(fp.read(25), b'\0' * 25)

    def test_gpt_image_partitions(self):
        image = Image(self.img, MiB(10), VolumeSchema.gpt)
        image.partition(offset=MiB(4), size=MiB(1), name='grub')
        self.assertEqual(len(image.disk.partitions), 1)
        image.partition(offset=MiB(5), size=MiB(4))
        self.assertEqual(len(image.disk.partitions), 2)
        image.set_parition_type(1, '21686148-6449-6E6F-744E-656564454649')
        image.set_parition_type(2, '0FC63DAF-8483-4772-8E79-3D69D8477DE4')
        # Use an external tool for checking the partition table to be sure
        # that it's indeed correct as suspected.
        disk_info = image.diagnostics()
        partitions = disk_info['partitiontable']
        # The device id is unpredictable.
        partitions.pop('id')
        # Newer sfdisk displays an additional field of 'sectorsize' that
        # we're not really interested in.
        partitions.pop('sectorsize', None)
        # The partition uuids as well.
        [p.pop('uuid') for p in partitions['partitions']]
        self.maxDiff = None
        self.assertEqual(partitions, {
            'label': 'gpt',
            'device': self.img,
            'unit': 'sectors',
            'firstlba': 34,
            'lastlba': 20446,
            'partitions': [{
                'node': '{}1'.format(self.img),
                'start': 8192,
                'size': 2048,
                'type': '21686148-6449-6E6F-744E-656564454649',
                'name': 'grub',
                }, {
                'node': '{}2'.format(self.img),
                'start': 10240,
                'size': 8192,
                'type': '0FC63DAF-8483-4772-8E79-3D69D8477DE4',
                }],
            })

    def test_write_value_at_offset(self):
        image = Image(self.img, MiB(2))
        image.write_value_at_offset(801, 130031)
        # Now open the path independently, seek to the given offset, and read
        # 4 bytes, then interpret it as a little-endian 32-bit integer.
        with open(image.path, 'rb') as fp:
            fp.seek(130031)
            # Unpack always returns a tuple, but there's only one item there.
            value, *ignore = unpack('<I', fp.read(4))
        self.assertEqual(value, 801)

    def test_write_value_at_offset_past_end(self):
        image = Image(self.img, 10000)
        self.assertRaises(ValueError, image.write_value_at_offset, 801, 130031)
        # And the file's size hasn't changed.
        self.assertEqual(os.path.getsize(self.img), 10000)

    def test_write_value_at_offsets_near_end(self):
        image = Image(self.img, 10000)
        # Attempt to write a bunch of values near the end of the file.  Since
        # the value will always be a 32-bit value, any positions farther out
        # than 4 bytes before the end will fail.
        results = set()
        for pos in range(9995, 10002):
            with suppress(ValueError):
                image.write_value_at_offset(801, pos)
                self.assertEqual(os.path.getsize(self.img), 10000)
                results.add(pos)
        self.assertEqual(results, {9995, 9996})

    def test_mbr_image_partitions(self):
        image = Image(self.img, MiB(2), VolumeSchema.mbr)
        # Create the first partition.
        image.partition(offset=image.sector(33),
                        size=image.sector(3000),
                        is_bootable=True)
        self.assertEqual(len(image.disk.partitions), 1)
        # Append the next one.
        image.partition(offset=image.sector(3033),
                        size=image.sector(1000))
        self.assertEqual(len(image.disk.partitions), 2)
        image.set_parition_type(1, '83')
        image.set_parition_type(2, 'dd')
        disk_info = image.diagnostics()
        partitions = disk_info['partitiontable']
        # The device id is unpredictable.
        partitions.pop('id')
        # XXX: In later versions of pyparted the partitiontable structure
        #  added a 'grain' entry that we're not really interested in.
        #  Remove it so we can have the tests working for all series.
        if 'grain' in partitions:
            partitions.pop('grain')
        # Newer sfdisk displays an additional field of 'sectorsize' that
        # we're not really interested in.
        partitions.pop('sectorsize', None)
        self.assertEqual(partitions, {
            'label': 'dos',
            'device': self.img,
            'unit': 'sectors',
            'partitions': [{
                'node': '{}1'.format(self.img),
                'start': 33,
                'size': 3000,
                'type': '83',
                'bootable': True,
                }, {
                'node': '{}2'.format(self.img),
                'start': 3033,
                'size': 1000,
                'type': 'dd',
                }],
            })

    def test_set_partition_type_gpt(self):
        image = Image(self.img, MiB(6), VolumeSchema.gpt)
        image.partition(offset=MiB(1), size=MiB(1))
        self.assertEqual(len(image.disk.partitions), 1)
        image.set_parition_type(1, '21686148-6449-6E6F-744E-656564454649')
        disk_info = image.diagnostics()
        self.assertEqual(disk_info['partitiontable']['partitions'][0]['type'],
                         '21686148-6449-6E6F-744E-656564454649')
        image.set_parition_type(1, '00000000-0000-0000-0000-0000DEADBEEF')
        disk_info = image.diagnostics()
        self.assertEqual(disk_info['partitiontable']['partitions'][0]['type'],
                         '00000000-0000-0000-0000-0000DEADBEEF')

    def test_set_partition_type_mbr(self):
        image = Image(self.img, MiB(6), VolumeSchema.mbr)
        image.partition(offset=MiB(1), size=MiB(1))
        self.assertEqual(len(image.disk.partitions), 1)
        image.set_parition_type(1, '83')
        disk_info = image.diagnostics()
        self.assertEqual(disk_info['partitiontable']['partitions'][0]['type'],
                         '83')
        image.set_parition_type(1, 'da')
        disk_info = image.diagnostics()
        self.assertEqual(disk_info['partitiontable']['partitions'][0]['type'],
                         'da')

    def test_set_partition_type_hybrid(self):
        image = Image(self.img, MiB(6), VolumeSchema.mbr)
        image.partition(offset=MiB(1), size=MiB(1))
        self.assertEqual(len(image.disk.partitions), 1)
        image.set_parition_type(
            1, ('83', '00000000-0000-0000-0000-0000DEADBEEF'))
        disk_info = image.diagnostics()
        self.assertEqual(disk_info['partitiontable']['partitions'][0]['type'],
                         '83')
        image.set_parition_type(
            1, ('da', '00000000-0000-0000-0000-0000DEADBEEF'))
        disk_info = image.diagnostics()
        self.assertEqual(disk_info['partitiontable']['partitions'][0]['type'],
                         'da')

    def test_sector_conversion(self):
        # For empty non-partitioned images we default to a 512 sector size.
        image = Image(self.img, MiB(1))
        self.assertEqual(image.sector(10), 5120)
        # In case of using partitioning, be sure we use the sector size as
        # returned by pyparted.
        image = Image(self.img, MiB(5), VolumeSchema.mbr)
        self.assertEqual(image.sector(10), 10 * image.device.sectorSize)

    def test_device_schema_required(self):
        # With no schema, the device cannot be partitioned.
        image = Image(self.img, MiB(1))
        self.assertRaises(TypeError, image.partition, 256, 512)

    def test_small_partition_size_and_offset(self):
        # LP: #1630709 - structure parts with size and offset < 1MB.
        image = Image(self.img, MiB(2), VolumeSchema.mbr)
        image.partition(offset=256, size=512)
        disk_info = image.diagnostics()
        # Even though the offset and size are set at 256 bytes and 512 bytes
        # respectively, the minimum granularity is one sector (i.e. 512
        # bytes).  The start and size returned by diagnostics() are in sector
        # units.
        self.assertEqual(
            disk_info['partitiontable']['partitions'][0]['start'],
            1)
        self.assertEqual(
            disk_info['partitiontable']['partitions'][0]['size'],
            1)
